# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import six

from gremlin_python.structure.io.graphsonV2d0 import (
    GraphSONReader as GraphSONReaderV2,
    GraphSONUtil as GraphSONUtil,  # no difference between v2 and v3
    VertexDeserializer as VertexDeserializerV2,
    VertexPropertyDeserializer as VertexPropertyDeserializerV2,
    PropertyDeserializer as PropertyDeserializerV2,
    EdgeDeserializer as EdgeDeserializerV2,
    PathDeserializer as PathDeserializerV2
)

from gremlin_python.structure.io.graphsonV3d0 import (
    GraphSONReader as GraphSONReaderV3,
    VertexDeserializer as VertexDeserializerV3,
    VertexPropertyDeserializer as VertexPropertyDeserializerV3,
    PropertyDeserializer as PropertyDeserializerV3,
    EdgeDeserializer as EdgeDeserializerV3,
    PathDeserializer as PathDeserializerV3
)

try:
    from gremlin_python.structure.io.graphsonV2d0 import (
        TraversalMetricsDeserializer as TraversalMetricsDeserializerV2,
        MetricsDeserializer as MetricsDeserializerV2
    )
    from gremlin_python.structure.io.graphsonV3d0 import (
        TraversalMetricsDeserializer as TraversalMetricsDeserializerV3,
        MetricsDeserializer as MetricsDeserializerV3
    )
except ImportError:
    TraversalMetricsDeserializerV2 = MetricsDeserializerV2 = None
    TraversalMetricsDeserializerV3 = MetricsDeserializerV3 = None

from cassandra.graph import (
    GraphSON2Serializer,
    GraphSON2Deserializer,
    GraphSON3Serializer,
    GraphSON3Deserializer
)
from cassandra.graph.graphson import UserTypeIO, TypeWrapperTypeIO
from cassandra.datastax.graph.fluent.predicates import GeoP, TextDistanceP
from cassandra.util import Distance


__all__ = ['GremlinGraphSONReader', 'GeoPSerializer', 'TextDistancePSerializer',
           'DistanceIO', 'gremlin_deserializers', 'deserializers', 'serializers',
           'GremlinGraphSONReaderV2', 'GremlinGraphSONReaderV3', 'dse_graphson2_serializers',
           'dse_graphson2_deserializers', 'dse_graphson3_serializers', 'dse_graphson3_deserializers',
           'gremlin_graphson2_deserializers', 'gremlin_graphson3_deserializers', 'GremlinUserTypeIO']


class _GremlinGraphSONTypeSerializer(object):
    TYPE_KEY = "@type"
    VALUE_KEY = "@value"
    serializer = None

    def __init__(self, serializer):
        self.serializer = serializer

    def dictify(self, v, writer):
        value = self.serializer.serialize(v, writer)
        if self.serializer is TypeWrapperTypeIO:
            graphson_base_type = v.type_io.graphson_base_type
            graphson_type = v.type_io.graphson_type
        else:
            graphson_base_type = self.serializer.graphson_base_type
            graphson_type = self.serializer.graphson_type

        if graphson_base_type is None:
            out = value
        else:
            out = {self.TYPE_KEY: graphson_type}
            if value is not None:
                out[self.VALUE_KEY] = value

        return out

    def definition(self, value, writer=None):
        return self.serializer.definition(value, writer)

    def get_specialized_serializer(self, value):
        ser = self.serializer.get_specialized_serializer(value)
        if ser is not self.serializer:
            return _GremlinGraphSONTypeSerializer(ser)
        return self


class _GremlinGraphSONTypeDeserializer(object):

    deserializer = None

    def __init__(self, deserializer):
        self.deserializer = deserializer

    def objectify(self, v, reader):
        return self.deserializer.deserialize(v, reader)


def _make_gremlin_graphson2_deserializer(graphson_type):
    return _GremlinGraphSONTypeDeserializer(
        GraphSON2Deserializer.get_deserializer(graphson_type.graphson_type)
    )


def _make_gremlin_graphson3_deserializer(graphson_type):
    return _GremlinGraphSONTypeDeserializer(
        GraphSON3Deserializer.get_deserializer(graphson_type.graphson_type)
    )


class _GremlinGraphSONReader(object):
    """Gremlin GraphSONReader Adapter, required to use gremlin types"""

    context = None

    def __init__(self, context, deserializer_map=None):
        self.context = context
        super(_GremlinGraphSONReader, self).__init__(deserializer_map)

    def deserialize(self, obj):
        return self.toObject(obj)


class GremlinGraphSONReaderV2(_GremlinGraphSONReader, GraphSONReaderV2):
    pass

# TODO remove next major
GremlinGraphSONReader = GremlinGraphSONReaderV2

class GremlinGraphSONReaderV3(_GremlinGraphSONReader, GraphSONReaderV3):
    pass


class GeoPSerializer(object):
    @classmethod
    def dictify(cls, p, writer):
        out = {
            "predicateType": "Geo",
            "predicate": p.operator,
            "value": [writer.toDict(p.value), writer.toDict(p.other)] if p.other is not None else writer.toDict(p.value)
        }
        return GraphSONUtil.typedValue("P", out, prefix='dse')


class TextDistancePSerializer(object):
    @classmethod
    def dictify(cls, p, writer):
        out = {
            "predicate": p.operator,
            "value": {
                'query': writer.toDict(p.value),
                'distance': writer.toDict(p.distance)
            }
        }
        return GraphSONUtil.typedValue("P", out)


class DistanceIO(object):
    @classmethod
    def dictify(cls, v, _):
        return GraphSONUtil.typedValue('Distance', six.text_type(v), prefix='dse')


GremlinUserTypeIO = _GremlinGraphSONTypeSerializer(UserTypeIO)

# GraphSON2
dse_graphson2_serializers = OrderedDict([
    (t, _GremlinGraphSONTypeSerializer(s))
    for t, s in six.iteritems(GraphSON2Serializer.get_type_definitions())
])

dse_graphson2_serializers.update(OrderedDict([
    (Distance, DistanceIO),
    (GeoP, GeoPSerializer),
    (TextDistanceP, TextDistancePSerializer)
]))

# TODO remove next major, this is just in case someone was using it
serializers = dse_graphson2_serializers

dse_graphson2_deserializers = {
    k: _make_gremlin_graphson2_deserializer(v)
    for k, v in six.iteritems(GraphSON2Deserializer.get_type_definitions())
}

dse_graphson2_deserializers.update({
    "dse:Distance": DistanceIO,
})

# TODO remove next major, this is just in case someone was using it
deserializers = dse_graphson2_deserializers

gremlin_graphson2_deserializers = dse_graphson2_deserializers.copy()
gremlin_graphson2_deserializers.update({
    'g:Vertex': VertexDeserializerV2,
    'g:VertexProperty': VertexPropertyDeserializerV2,
    'g:Edge': EdgeDeserializerV2,
    'g:Property': PropertyDeserializerV2,
    'g:Path': PathDeserializerV2
})

if TraversalMetricsDeserializerV2:
    gremlin_graphson2_deserializers.update({
        'g:TraversalMetrics': TraversalMetricsDeserializerV2,
        'g:lMetrics': MetricsDeserializerV2
    })

# TODO remove next major, this is just in case someone was using it
gremlin_deserializers = gremlin_graphson2_deserializers

# GraphSON3
dse_graphson3_serializers = OrderedDict([
    (t, _GremlinGraphSONTypeSerializer(s))
    for t, s in six.iteritems(GraphSON3Serializer.get_type_definitions())
])

dse_graphson3_serializers.update(OrderedDict([
    (Distance, DistanceIO),
    (GeoP, GeoPSerializer),
    (TextDistanceP, TextDistancePSerializer)
]))

dse_graphson3_deserializers = {
    k: _make_gremlin_graphson3_deserializer(v)
    for k, v in six.iteritems(GraphSON3Deserializer.get_type_definitions())
}

dse_graphson3_deserializers.update({
    "dse:Distance": DistanceIO
})

gremlin_graphson3_deserializers = dse_graphson3_deserializers.copy()
gremlin_graphson3_deserializers.update({
    'g:Vertex': VertexDeserializerV3,
    'g:VertexProperty': VertexPropertyDeserializerV3,
    'g:Edge': EdgeDeserializerV3,
    'g:Property': PropertyDeserializerV3,
    'g:Path': PathDeserializerV3
})

if TraversalMetricsDeserializerV3:
    gremlin_graphson3_deserializers.update({
        'g:TraversalMetrics': TraversalMetricsDeserializerV3,
        'g:Metrics': MetricsDeserializerV3
    })



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import logging
import six
import threading

from cassandra.cluster import Cluster, _ConfigMode, _NOT_SET, NoHostAvailable, UserTypeDoesNotExist, ConsistencyLevel
from cassandra.query import SimpleStatement, dict_factory

from cassandra.cqlengine import CQLEngineException
from cassandra.cqlengine.statements import BaseCQLStatement


log = logging.getLogger(__name__)

NOT_SET = _NOT_SET  # required for passing timeout to Session.execute

cluster = None
session = None

# connections registry
DEFAULT_CONNECTION = object()
_connections = {}

# Because type models may be registered before a connection is present,
# and because sessions may be replaced, we must register UDTs here, in order
# to have them registered when a new session is established.
udt_by_keyspace = defaultdict(dict)


def format_log_context(msg, connection=None, keyspace=None):
    """Format log message to add keyspace and connection context"""
    connection_info = connection or 'DEFAULT_CONNECTION'

    if keyspace:
        msg = '[Connection: {0}, Keyspace: {1}] {2}'.format(connection_info, keyspace, msg)
    else:
        msg = '[Connection: {0}] {1}'.format(connection_info, msg)
    return msg


class UndefinedKeyspaceException(CQLEngineException):
    pass


class Connection(object):
    """CQLEngine Connection"""

    name = None
    hosts = None

    consistency = None
    retry_connect = False
    lazy_connect = False
    lazy_connect_lock = None
    cluster_options = None

    cluster = None
    session = None

    def __init__(self, name, hosts, consistency=None,
                 lazy_connect=False, retry_connect=False, cluster_options=None):
        self.hosts = hosts
        self.name = name
        self.consistency = consistency
        self.lazy_connect = lazy_connect
        self.retry_connect = retry_connect
        self.cluster_options = cluster_options if cluster_options else {}
        self.lazy_connect_lock = threading.RLock()

    @classmethod
    def from_session(cls, name, session):
        instance = cls(name=name, hosts=session.hosts)
        instance.cluster, instance.session = session.cluster, session
        instance.setup_session()
        return instance

    def setup(self):
        """Setup the connection"""
        global cluster, session

        if 'username' in self.cluster_options or 'password' in self.cluster_options:
            raise CQLEngineException("Username & Password are now handled by using the native driver's auth_provider")

        if self.lazy_connect:
            return

        self.cluster = Cluster(self.hosts, **self.cluster_options)
        try:
            self.session = self.cluster.connect()
            log.debug(format_log_context("connection initialized with internally created session", connection=self.name))
        except NoHostAvailable:
            if self.retry_connect:
                log.warning(format_log_context("connect failed, setting up for re-attempt on first use", connection=self.name))
                self.lazy_connect = True
            raise

        if DEFAULT_CONNECTION in _connections and _connections[DEFAULT_CONNECTION] == self:
            cluster = _connections[DEFAULT_CONNECTION].cluster
            session = _connections[DEFAULT_CONNECTION].session

        self.setup_session()

    def setup_session(self):
        if self.cluster._config_mode == _ConfigMode.PROFILES:
            self.cluster.profile_manager.default.row_factory = dict_factory
            if self.consistency is not None:
                self.cluster.profile_manager.default.consistency_level = self.consistency
        else:
            self.session.row_factory = dict_factory
            if self.consistency is not None:
                self.session.default_consistency_level = self.consistency
        enc = self.session.encoder
        enc.mapping[tuple] = enc.cql_encode_tuple
        _register_known_types(self.session.cluster)

    def handle_lazy_connect(self):

        # if lazy_connect is False, it means the cluster is setup and ready
        # No need to acquire the lock
        if not self.lazy_connect:
            return

        with self.lazy_connect_lock:
            # lazy_connect might have been set to False by another thread while waiting the lock
            # In this case, do nothing.
            if self.lazy_connect:
                log.debug(format_log_context("Lazy connect enabled", connection=self.name))
                self.lazy_connect = False
                self.setup()


def register_connection(name, hosts=None, consistency=None, lazy_connect=False,
                        retry_connect=False, cluster_options=None, default=False,
                        session=None):
    """
    Add a connection to the connection registry. ``hosts`` and ``session`` are
    mutually exclusive, and ``consistency``, ``lazy_connect``,
    ``retry_connect``, and ``cluster_options`` only work with ``hosts``. Using
    ``hosts`` will create a new :class:`cassandra.cluster.Cluster` and
    :class:`cassandra.cluster.Session`.

    :param list hosts: list of hosts, (``contact_points`` for :class:`cassandra.cluster.Cluster`).
    :param int consistency: The default :class:`~.ConsistencyLevel` for the
        registered connection's new session. Default is the same as
        :attr:`.Session.default_consistency_level`. For use with ``hosts`` only;
        will fail when used with ``session``.
    :param bool lazy_connect: True if should not connect until first use. For
        use with ``hosts`` only; will fail when used with ``session``.
    :param bool retry_connect: True if we should retry to connect even if there
        was a connection failure initially. For use with ``hosts`` only; will
        fail when used with ``session``.
    :param dict cluster_options: A dict of options to be used as keyword
        arguments to :class:`cassandra.cluster.Cluster`. For use with ``hosts``
        only; will fail when used with ``session``.
    :param bool default: If True, set the new connection as the cqlengine
        default
    :param Session session: A :class:`cassandra.cluster.Session` to be used in
        the created connection.
    """

    if name in _connections:
        log.warning("Registering connection '{0}' when it already exists.".format(name))

    if session is not None:
        invalid_config_args = (hosts is not None or
                               consistency is not None or
                               lazy_connect is not False or
                               retry_connect is not False or
                               cluster_options is not None)
        if invalid_config_args:
            raise CQLEngineException(
                "Session configuration arguments and 'session' argument are mutually exclusive"
            )
        conn = Connection.from_session(name, session=session)
    else:  # use hosts argument
        conn = Connection(
            name, hosts=hosts,
            consistency=consistency, lazy_connect=lazy_connect,
            retry_connect=retry_connect, cluster_options=cluster_options
        )
        conn.setup()

    _connections[name] = conn

    if default:
        set_default_connection(name)

    return conn


def unregister_connection(name):
    global cluster, session

    if name not in _connections:
        return

    if DEFAULT_CONNECTION in _connections and _connections[name] == _connections[DEFAULT_CONNECTION]:
        del _connections[DEFAULT_CONNECTION]
        cluster = None
        session = None

    conn = _connections[name]
    if conn.cluster:
        conn.cluster.shutdown()
    del _connections[name]
    log.debug("Connection '{0}' has been removed from the registry.".format(name))


def set_default_connection(name):
    global cluster, session

    if name not in _connections:
        raise CQLEngineException("Connection '{0}' doesn't exist.".format(name))

    log.debug("Connection '{0}' has been set as default.".format(name))
    _connections[DEFAULT_CONNECTION] = _connections[name]
    cluster = _connections[name].cluster
    session = _connections[name].session


def get_connection(name=None):

    if not name:
        name = DEFAULT_CONNECTION

    if name not in _connections:
        raise CQLEngineException("Connection name '{0}' doesn't exist in the registry.".format(name))

    conn = _connections[name]
    conn.handle_lazy_connect()

    return conn


def default():
    """
    Configures the default connection to localhost, using the driver defaults
    (except for row_factory)
    """

    try:
        conn = get_connection()
        if conn.session:
            log.warning("configuring new default connection for cqlengine when one was already set")
    except:
        pass

    register_connection('default', hosts=None, default=True)

    log.debug("cqlengine connection initialized with default session to localhost")


def set_session(s):
    """
    Configures the default connection with a preexisting :class:`cassandra.cluster.Session`

    Note: the mapper presently requires a Session :attr:`~.row_factory` set to ``dict_factory``.
    This may be relaxed in the future
    """

    try:
        conn = get_connection()
    except CQLEngineException:
        # no default connection set; initalize one
        register_connection('default', session=s, default=True)
        conn = get_connection()
    else:
        if conn.session:
            log.warning("configuring new default session for cqlengine when one was already set")

    if not any([
        s.cluster.profile_manager.default.row_factory is dict_factory and s.cluster._config_mode in [_ConfigMode.PROFILES, _ConfigMode.UNCOMMITTED],
        s.row_factory is dict_factory and s.cluster._config_mode in [_ConfigMode.LEGACY, _ConfigMode.UNCOMMITTED],
    ]):
        raise CQLEngineException("Failed to initialize: row_factory must be 'dict_factory'")

    conn.session = s
    conn.cluster = s.cluster

    # Set default keyspace from given session's keyspace
    if conn.session.keyspace:
        from cassandra.cqlengine import models
        models.DEFAULT_KEYSPACE = conn.session.keyspace

    conn.setup_session()

    log.debug("cqlengine default connection initialized with %s", s)


def setup(
        hosts,
        default_keyspace,
        consistency=None,
        lazy_connect=False,
        retry_connect=False,
        **kwargs):
    """
    Setup a the driver connection used by the mapper

    :param list hosts: list of hosts, (``contact_points`` for :class:`cassandra.cluster.Cluster`)
    :param str default_keyspace: The default keyspace to use
    :param int consistency: The global default :class:`~.ConsistencyLevel` - default is the same as :attr:`.Session.default_consistency_level`
    :param bool lazy_connect: True if should not connect until first use
    :param bool retry_connect: True if we should retry to connect even if there was a connection failure initially
    :param \*\*kwargs: Pass-through keyword arguments for :class:`cassandra.cluster.Cluster`
    """

    from cassandra.cqlengine import models
    models.DEFAULT_KEYSPACE = default_keyspace

    register_connection('default', hosts=hosts, consistency=consistency, lazy_connect=lazy_connect,
                        retry_connect=retry_connect, cluster_options=kwargs, default=True)


def execute(query, params=None, consistency_level=None, timeout=NOT_SET, connection=None):

    conn = get_connection(connection)

    if not conn.session:
        raise CQLEngineException("It is required to setup() cqlengine before executing queries")

    if isinstance(query, SimpleStatement):
        pass  #
    elif isinstance(query, BaseCQLStatement):
        params = query.get_context()
        query = SimpleStatement(str(query), consistency_level=consistency_level, fetch_size=query.fetch_size)
    elif isinstance(query, six.string_types):
        query = SimpleStatement(query, consistency_level=consistency_level)
    log.debug(format_log_context('Query: {}, Params: {}'.format(query.query_string, params), connection=connection))

    result = conn.session.execute(query, params, timeout=timeout)

    return result


def get_session(connection=None):
    conn = get_connection(connection)
    return conn.session


def get_cluster(connection=None):
    conn = get_connection(connection)
    if not conn.cluster:
        raise CQLEngineException("%s.cluster is not configured. Call one of the setup or default functions first." % __name__)
    return conn.cluster


def register_udt(keyspace, type_name, klass, connection=None):
    udt_by_keyspace[keyspace][type_name] = klass

    try:
        cluster = get_cluster(connection)
    except CQLEngineException:
        cluster = None

    if cluster:
        try:
            cluster.register_user_type(keyspace, type_name, klass)
        except UserTypeDoesNotExist:
            pass  # new types are covered in management sync functions


def _register_known_types(cluster):
    from cassandra.cqlengine import models
    for ks_name, name_type_map in udt_by_keyspace.items():
        for type_name, klass in name_type_map.items():
            try:
                cluster.register_user_type(ks_name or models.DEFAULT_KEYSPACE, type_name, klass)
            except UserTypeDoesNotExist:
                pass  # new types are covered in management sync functions



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from collections import OrderedDict
from warnings import warn

from cassandra.datastax.insights.util import namespace

_NOT_SET = object()


def _default_serializer_for_object(obj, policy):
    # the insights server expects an 'options' dict for policy
    # objects, but not for other objects
    if policy:
        return {'type': obj.__class__.__name__,
                'namespace': namespace(obj.__class__),
                'options': {}}
    else:
        return {'type': obj.__class__.__name__,
                'namespace': namespace(obj.__class__)}


class InsightsSerializerRegistry(object):

    initialized = False

    def __init__(self, mapping_dict=None):
        mapping_dict = mapping_dict or {}
        class_order = self._class_topological_sort(mapping_dict)
        self._mapping_dict = OrderedDict(
            ((cls, mapping_dict[cls]) for cls in class_order)
        )

    def serialize(self, obj, policy=False, default=_NOT_SET, cls=None):
        try:
            return self._get_serializer(cls if cls is not None else obj.__class__)(obj)
        except Exception:
            if default is _NOT_SET:
                result = _default_serializer_for_object(obj, policy)
            else:
                result = default

            return result

    def _get_serializer(self, cls):
        try:
            return self._mapping_dict[cls]
        except KeyError:
            for registered_cls, serializer in six.iteritems(self._mapping_dict):
                if issubclass(cls, registered_cls):
                    return self._mapping_dict[registered_cls]
        raise ValueError

    def register(self, cls, serializer):
        self._mapping_dict[cls] = serializer
        self._mapping_dict = OrderedDict(
            ((cls, self._mapping_dict[cls])
             for cls in self._class_topological_sort(self._mapping_dict))
        )

    def register_serializer_for(self, cls):
        """
        Parameterized registration helper decorator. Given a class `cls`,
        produces a function that registers the decorated function as a
        serializer for it.
        """
        def decorator(serializer):
            self.register(cls, serializer)
            return serializer

        return decorator

    @staticmethod
    def _class_topological_sort(classes):
        """
        A simple topological sort for classes. Takes an iterable of class objects
        and returns a list A of those classes, ordered such that A[X] is never a
        superclass of A[Y] for X < Y.

        This is an inefficient sort, but that's ok because classes are infrequently
        registered. It's more important that this be maintainable than fast.

        We can't use `.sort()` or `sorted()` with a custom `key` -- those assume
        a total ordering, which we don't have.
        """
        unsorted, sorted_ = list(classes), []
        while unsorted:
            head, tail = unsorted[0], unsorted[1:]

            # if head has no subclasses remaining, it can safely go in the list
            if not any(issubclass(x, head) for x in tail):
                sorted_.append(head)
            else:
                # move to the back -- head has to wait until all its subclasses
                # are sorted into the list
                tail.append(head)

            unsorted = tail

        # check that sort is valid
        for i, head in enumerate(sorted_):
            for after_head_value in sorted_[(i + 1):]:
                if issubclass(after_head_value, head):
                    warn('Sorting classes produced an invalid ordering.\n'
                         'In:  {classes}\n'
                         'Out: {sorted_}'.format(classes=classes, sorted_=sorted_))
        return sorted_


insights_registry = InsightsSerializerRegistry()



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import json
import sys
import tempfile
import shutil
import six
from six.moves.urllib.request import urlopen

_HAS_SSL = True
try:
    from ssl import SSLContext, PROTOCOL_TLS, CERT_REQUIRED
except:
    _HAS_SSL = False

from zipfile import ZipFile

# 2.7 vs 3.x
try:
    from zipfile import BadZipFile
except:
    from zipfile import BadZipfile as BadZipFile

from cassandra import DriverException

log = logging.getLogger(__name__)

__all__ = ['get_cloud_config']

DATASTAX_CLOUD_PRODUCT_TYPE = "DATASTAX_APOLLO"


class CloudConfig(object):

    username = None
    password = None
    host = None
    port = None
    keyspace = None
    local_dc = None
    ssl_context = None

    sni_host = None
    sni_port = None
    host_ids = None

    @classmethod
    def from_dict(cls, d):
        c = cls()

        c.port = d.get('port', None)
        try:
            c.port = int(d['port'])
        except:
            pass

        c.username = d.get('username', None)
        c.password = d.get('password', None)
        c.host = d.get('host', None)
        c.keyspace = d.get('keyspace', None)
        c.local_dc = d.get('localDC', None)

        return c


def get_cloud_config(cloud_config, create_pyopenssl_context=False):
    if not _HAS_SSL:
        raise DriverException("A Python installation with SSL is required to connect to a cloud cluster.")

    if 'secure_connect_bundle' not in cloud_config:
        raise ValueError("The cloud config doesn't have a secure_connect_bundle specified.")

    try:
        config = read_cloud_config_from_zip(cloud_config, create_pyopenssl_context)
    except BadZipFile:
        raise ValueError("Unable to open the zip file for the cloud config. Check your secure connect bundle.")

    config = read_metadata_info(config, cloud_config)
    if create_pyopenssl_context:
        config.ssl_context = config.pyopenssl_context
    return config


def read_cloud_config_from_zip(cloud_config, create_pyopenssl_context):
    secure_bundle = cloud_config['secure_connect_bundle']
    use_default_tempdir = cloud_config.get('use_default_tempdir', None)
    with ZipFile(secure_bundle) as zipfile:
        base_dir = tempfile.gettempdir() if use_default_tempdir else os.path.dirname(secure_bundle)
        tmp_dir = tempfile.mkdtemp(dir=base_dir)
        try:
            zipfile.extractall(path=tmp_dir)
            return parse_cloud_config(os.path.join(tmp_dir, 'config.json'), cloud_config, create_pyopenssl_context)
        finally:
            shutil.rmtree(tmp_dir)


def parse_cloud_config(path, cloud_config, create_pyopenssl_context):
    with open(path, 'r') as stream:
        data = json.load(stream)

    config = CloudConfig.from_dict(data)
    config_dir = os.path.dirname(path)

    if 'ssl_context' in cloud_config:
        config.ssl_context = cloud_config['ssl_context']
    else:
        # Load the ssl_context before we delete the temporary directory
        ca_cert_location = os.path.join(config_dir, 'ca.crt')
        cert_location = os.path.join(config_dir, 'cert')
        key_location = os.path.join(config_dir, 'key')
        # Regardless of if we create a pyopenssl context, we still need the builtin one
        # to connect to the metadata service
        config.ssl_context = _ssl_context_from_cert(ca_cert_location, cert_location, key_location)
        if create_pyopenssl_context:
            config.pyopenssl_context = _pyopenssl_context_from_cert(ca_cert_location, cert_location, key_location)

    return config


def read_metadata_info(config, cloud_config):
    url = "https://{}:{}/metadata".format(config.host, config.port)
    timeout = cloud_config['connect_timeout'] if 'connect_timeout' in cloud_config else 5
    try:
        response = urlopen(url, context=config.ssl_context, timeout=timeout)
    except Exception as e:
        log.exception(e)
        raise DriverException("Unable to connect to the metadata service at %s. "
                              "Check the cluster status in the cloud console. " % url)

    if response.code != 200:
        raise DriverException(("Error while fetching the metadata at: %s. "
                               "The service returned error code %d." % (url, response.code)))
    return parse_metadata_info(config, response.read().decode('utf-8'))


def parse_metadata_info(config, http_data):
    try:
        data = json.loads(http_data)
    except:
        msg = "Failed to load cluster metadata"
        raise DriverException(msg)

    contact_info = data['contact_info']
    config.local_dc = contact_info['local_dc']

    proxy_info = contact_info['sni_proxy_address'].split(':')
    config.sni_host = proxy_info[0]
    try:
        config.sni_port = int(proxy_info[1])
    except:
        config.sni_port = 9042

    config.host_ids = [host_id for host_id in contact_info['contact_points']]

    return config


def _ssl_context_from_cert(ca_cert_location, cert_location, key_location):
    ssl_context = SSLContext(PROTOCOL_TLS)
    ssl_context.load_verify_locations(ca_cert_location)
    ssl_context.verify_mode = CERT_REQUIRED
    ssl_context.load_cert_chain(certfile=cert_location, keyfile=key_location)

    return ssl_context


def _pyopenssl_context_from_cert(ca_cert_location, cert_location, key_location):
    try:
        from OpenSSL import SSL
    except ImportError as e:
        six.reraise(
            ImportError,
            ImportError("PyOpenSSL must be installed to connect to Astra with the Eventlet or Twisted event loops"),
            sys.exc_info()[2]
        )
    ssl_context = SSL.Context(SSL.TLSv1_METHOD)
    ssl_context.set_verify(SSL.VERIFY_PEER, callback=lambda _1, _2, _3, _4, ok: ok)
    ssl_context.use_certificate_file(cert_location)
    ssl_context.use_privatekey_file(key_location)
    ssl_context.load_verify_locations(ca_cert_location)

    return ssl_context